#define WIN32_LEAN_AND_MEAN
#define _CRT_SECURE_NO_DEPRECATE
#include <windows.h>
#include <commctrl.h>

#include <stdlib.h>
#include <stdio.h>

#include "SignalGenerator.h"
#include "comport.h"

#define CONTROL 0
#define ACCU 1
#define ACCU_INCREMENT 2
#define SQUARE_WAVE_THRESHOLD 3
#define SIN_REGISTER 4
#define COS_REGISTER 5
#define RANDOM_REGISTER 6
#define INPUT_REGISTER 7
#define OUTPUT_REGISTER 8
#define VERSION_REGISTER 15

long long registers[15] = { 0 };
long long transferedRegisters[15] = { 0 };

enum Waveform {
	WAVEFORM_OFF,
	WAVEFORM_SINE,
	WAVEFORM_SQUARE_WAVE,
	WAVEFORM_NOISE,
	WAVEFORM_SAWTOOTH
};

// data
int comPort = 1;
bool connected = false;
bool syncAll = false;
Waveform waveform = WAVEFORM_OFF;
bool inverseSawtooth = false;
double frequency = 1000.0l;
double frequencyStart = 0.1l;
double frequencyEnd = 100000.0l;
double periodStart = 10e-6;
double periodEnd = 10.0l;
int digitalInputBits = 0;
int digitalOutputBits = 0;
double dutyCycle = 50.0l;

// GUI
bool inData2gui = false;
double sliderRange = 200.0l;
HWND comPortWnd, connectWnd, disconnectWnd;
HWND offWnd, sineWnd, squareWaveWnd, noiseWnd, sawtoothWnd;
HWND inverseSawtoothWnd;
HWND frequencyStartWnd, frequencyWnd, frequencyEndWnd, frequencySliderWnd;
HWND periodStartWnd, periodWnd, periodEndWnd, periodSliderWnd;
HWND input0Wnd, input1Wnd, input2Wnd, input3Wnd;
HWND input4Wnd, input5Wnd, input6Wnd, input7Wnd;
HWND output0Wnd, output1Wnd, output2Wnd, output3Wnd;
HWND output4Wnd, output5Wnd, output6Wnd, output7Wnd;
HWND dutyCycleWnd, dutyCycleSliderWnd, lowTimeWnd, highTimeWnd;

void setText(HWND control, const char* text)
{
	SendMessage(control, (UINT) WM_SETTEXT, (WPARAM) 0, (LPARAM) text);
}

void setIntField(HWND control, int value)
{
	char buf[256];
	sprintf(buf, "%i", value);
	SendMessage(control, (UINT) WM_SETTEXT, (WPARAM) 0, (LPARAM) buf);
}

void setDoubleField(HWND control, double value)
{
	char buf[256];
	sprintf(buf, "%g", value);
	SendMessage(control, (UINT) WM_SETTEXT, (WPARAM) 0, (LPARAM) buf);
}

void setButtonChecked(HWND control, bool checked)
{
	SendMessage(control, (UINT) BM_SETCHECK, (WPARAM) (checked ? BST_CHECKED : BST_UNCHECKED), (LPARAM) 0);
}

bool isButtonChecked(HWND control)
{
	return SendMessage(control, (UINT) BM_GETCHECK, (WPARAM) 0, (LPARAM) 0) == TRUE;
}

void setEnabled(HWND control, bool enabled)
{
	EnableWindow(control, enabled ? TRUE : FALSE);
}

void setSliderValue(HWND control, double start, double end, double value)
{
	double delta = end - start;
	double dpos = (value - start) / delta * sliderRange;
	if (dpos < 0) dpos = 0;
	if (dpos > sliderRange) dpos = sliderRange;

	int pos = (int) dpos;
	int min = 0;
	int max = (int) sliderRange;

	SendMessage(control, TBM_SETRANGE, (WPARAM) TRUE, (LPARAM) MAKELONG(min, max));
    SendMessage(control, TBM_SETPAGESIZE, 0, (LPARAM) max / 10);
    SendMessage(control, TBM_SETSEL, (WPARAM) FALSE, (LPARAM) MAKELONG(min, max));
    SendMessage(control, TBM_SETPOS, (WPARAM) TRUE, (LPARAM) pos);
}

double getDoubleField(HWND control, double defaultValue)
{
	char buf[256];
	_CRT_DOUBLE result;
	SendMessage(control, (UINT) WM_GETTEXT, (WPARAM) 256, (LPARAM) buf);
	int retval = _atodbl(&result, buf);
	if (retval != 0 || result.x == 0) {
		return defaultValue;
	} else {
		return result.x;
	}
}

void updateInput()
{
	setButtonChecked(input0Wnd, (digitalInputBits & 0x01) > 0);
	setButtonChecked(input1Wnd, (digitalInputBits & 0x02) > 0);
	setButtonChecked(input2Wnd, (digitalInputBits & 0x04) > 0);
	setButtonChecked(input3Wnd, (digitalInputBits & 0x08) > 0);
	setButtonChecked(input4Wnd, (digitalInputBits & 0x10) > 0);
	setButtonChecked(input5Wnd, (digitalInputBits & 0x20) > 0);
	setButtonChecked(input6Wnd, (digitalInputBits & 0x40) > 0);
	setButtonChecked(input7Wnd, (digitalInputBits & 0x80) > 0);
}

void data2gui(HWND dontSet)
{
	// don't call it recursive
	if (!inData2gui) {
		inData2gui = true;
		setIntField(comPortWnd, comPort);
		setEnabled(connectWnd, !connected);
		setEnabled(disconnectWnd, connected);
		setButtonChecked(offWnd, false);
		setButtonChecked(sineWnd, false);
		setButtonChecked(squareWaveWnd, false);
		setButtonChecked(noiseWnd, false);
		setButtonChecked(sawtoothWnd, false);
		switch (waveform) {
			case WAVEFORM_OFF: setButtonChecked(offWnd, true); break;
			case WAVEFORM_SINE: setButtonChecked(sineWnd, true); break;
			case WAVEFORM_SQUARE_WAVE: setButtonChecked(squareWaveWnd, true); break;
			case WAVEFORM_NOISE: setButtonChecked(noiseWnd, true); break;
			case WAVEFORM_SAWTOOTH: setButtonChecked(sawtoothWnd, true); break;
		}
		setButtonChecked(inverseSawtoothWnd, inverseSawtooth);
		if (dontSet != frequencyStartWnd) setDoubleField(frequencyStartWnd, frequencyStart);
		if (frequency < 1e-30) frequency = 1e-30;
		if (dontSet != frequencyWnd) setDoubleField(frequencyWnd, frequency);
		if (dontSet != frequencyEndWnd) setDoubleField(frequencyEndWnd, frequencyEnd);
		if (dontSet != frequencySliderWnd) setSliderValue(frequencySliderWnd, frequencyStart, frequencyEnd, frequency);
		if (dontSet != periodStartWnd) setDoubleField(periodStartWnd, periodStart);
		double period = 1.0l / frequency;
		double lowTime = period * dutyCycle / 100.0l;
		double highTime = period * (100.0l - dutyCycle) / 100.0l;
		if (dontSet != periodWnd) setDoubleField(periodWnd, period);
		if (dontSet != periodEndWnd) setDoubleField(periodEndWnd, periodEnd);
		setSliderValue(periodSliderWnd, periodStart, periodEnd, 1.0l/frequency);
		updateInput();
		if (dontSet != dutyCycleWnd) setDoubleField(dutyCycleWnd, dutyCycle);
		setSliderValue(dutyCycleSliderWnd, 0.0l, 100.0l, dutyCycle);
		if (dontSet != lowTimeWnd) setDoubleField(lowTimeWnd, lowTime);
		if (dontSet != highTimeWnd) setDoubleField(highTimeWnd, highTime);

		inData2gui = false;
	}
}

void onFrequencySliderChanged()
{
	int pos = (int) SendMessage(frequencySliderWnd, TBM_GETPOS, 0, 0);
	double delta = frequencyEnd - frequencyStart;
	frequency = ((double)pos) / sliderRange * delta + frequencyStart;
	data2gui(frequencySliderWnd);
}

void onPeriodSliderChanged()
{
	int pos = (int) SendMessage(periodSliderWnd, TBM_GETPOS, 0, 0);
	double delta = periodEnd - periodStart;
	frequency = 1.0l / (((double)pos) / sliderRange * delta + periodStart);
	data2gui(periodSliderWnd);
}

void onDutyCycleSliderChanged()
{
	int pos = (int) SendMessage(dutyCycleSliderWnd, TBM_GETPOS, 0, 0);
	dutyCycle = ((double)pos) / sliderRange * 100.0l;
	data2gui(dutyCycleSliderWnd);
}

void onFrequencyStartChanged()
{
	frequencyStart = getDoubleField(frequencyStartWnd, frequencyStart);
	data2gui(frequencyStartWnd);
}

void onFrequencyChanged()
{
	frequency = getDoubleField(frequencyWnd, frequency);
	data2gui(frequencyWnd);
}

void onFrequencyEndChanged()
{
	frequencyEnd = getDoubleField(frequencyEndWnd, frequencyEnd);
	data2gui(frequencyEndWnd);
}

void onPeriodStartChanged()
{
	periodStart = getDoubleField(periodStartWnd, periodStart);
	data2gui(periodStartWnd);
}

void onPeriodChanged()
{
	double result = getDoubleField(periodWnd, 1.0l / frequency);
	if (result < 1e-30) result = 1e-30;
	frequency = 1.0l / result;
	data2gui(periodWnd);
}

void onPeriodEndChanged()
{
	periodEnd = getDoubleField(periodEndWnd, periodEnd);
	data2gui(periodEndWnd);
}

void onDutyCycleChanged()
{
	dutyCycle = getDoubleField(dutyCycleWnd, dutyCycle);
	data2gui(dutyCycleWnd);
}

void onLowTimeChanged()
{
	double period = 1.0l / frequency;
	double lowTime = period * dutyCycle / 100.0l;
	double highTime = period * (100.0l - dutyCycle) / 100.0l;
	lowTime = getDoubleField(lowTimeWnd, lowTime);
	period = lowTime + highTime;
	frequency = 1.0l / period;
	dutyCycle = lowTime / period * 100.0l;
	data2gui(lowTimeWnd);
}

void onHighTimeChanged()
{
	double period = 1.0l / frequency;
	double lowTime = period * dutyCycle / 100.0l;
	double highTime = period * (100.0l - dutyCycle) / 100.0l;
	highTime = getDoubleField(highTimeWnd, highTime);
	period = lowTime + highTime;
	frequency = 1.0l / period;
	dutyCycle = lowTime / period * 100.0l;
	data2gui(lowTimeWnd);
}

void syncDevice()
{
	// read last input register
	string text = comRead();
	if (text.length() > 0) {
		long long in;
		sscanf(text.c_str(), "%llx", &in);
		int input = int(in & 0xff);
		if (input != digitalInputBits) {
			digitalInputBits = input;
			updateInput();
		}
	}

	// update registers
	registers[CONTROL] = (int) waveform;
	registers[ACCU_INCREMENT] = (long long) (((double) 0x10000000000) / 1.0e6l * frequency);
	if (inverseSawtooth && waveform == WAVEFORM_SAWTOOTH) registers[ACCU_INCREMENT] = 0x10000000000 - registers[ACCU_INCREMENT];
	registers[SQUARE_WAVE_THRESHOLD] = (long long) (((double) 0x10000000000) * dutyCycle / 100.0l);
	registers[OUTPUT_REGISTER] = digitalOutputBits;

	// sync changed registers
	char buf[256];
	for (int i = 0; i < 15; i++) {
		if (transferedRegisters[i] != registers[i] || syncAll) {
			sprintf(buf, "%i=%llx\n", i, registers[i]);
			transferedRegisters[i] = registers[i];
			comWrite(buf);
		}
	}
	syncAll = false;

	// schedule read of input register
	sprintf(buf, "%i\n", INPUT_REGISTER);
	comWrite(buf);
}

INT_PTR CALLBACK mainDialog(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	HWND hwnd;
	switch (message)
	{
		case WM_INITDIALOG:
			// get window handles for all controls
			comPortWnd = GetDlgItem(hDlg, IDC_COM_PORT);
			connectWnd = GetDlgItem(hDlg, IDC_CONNECT);
			disconnectWnd = GetDlgItem(hDlg, IDC_DISCONNECT);
			offWnd = GetDlgItem(hDlg, IDC_OFF);
			sineWnd = GetDlgItem(hDlg, IDC_SINE);
			squareWaveWnd = GetDlgItem(hDlg, IDC_SQUARE_WAVE);
			noiseWnd = GetDlgItem(hDlg, IDC_NOISE);
			sawtoothWnd = GetDlgItem(hDlg, IDC_SAWTOOTH);
			inverseSawtoothWnd = GetDlgItem(hDlg, IDC_INVERSE_SAWTOOTH);
			frequencyStartWnd = GetDlgItem(hDlg, IDC_FREQUENCY_START);
			frequencyWnd = GetDlgItem(hDlg, IDC_FREQUENCY);
			frequencyEndWnd = GetDlgItem(hDlg, IDC_FREQUENCY_END);
			frequencySliderWnd = GetDlgItem(hDlg, IDC_FREQUENCY_SLIDER);
			periodStartWnd = GetDlgItem(hDlg, IDC_PERIOD_START);
			periodWnd = GetDlgItem(hDlg, IDC_PERIOD);
			periodEndWnd = GetDlgItem(hDlg, IDC_PERIOD_END);
			periodSliderWnd = GetDlgItem(hDlg, IDC_PERIOD_SLIDER);
			input0Wnd = GetDlgItem(hDlg, IDC_INPUT0);
			input1Wnd = GetDlgItem(hDlg, IDC_INPUT1);
			input2Wnd = GetDlgItem(hDlg, IDC_INPUT2);
			input3Wnd = GetDlgItem(hDlg, IDC_INPUT3);
			input4Wnd = GetDlgItem(hDlg, IDC_INPUT4);
			input5Wnd = GetDlgItem(hDlg, IDC_INPUT5);
			input6Wnd = GetDlgItem(hDlg, IDC_INPUT6);
			input7Wnd = GetDlgItem(hDlg, IDC_INPUT7);
			output0Wnd = GetDlgItem(hDlg, IDC_OUTPUT0);
			output1Wnd = GetDlgItem(hDlg, IDC_OUTPUT1);
			output2Wnd = GetDlgItem(hDlg, IDC_OUTPUT2);
			output3Wnd = GetDlgItem(hDlg, IDC_OUTPUT3);
			output4Wnd = GetDlgItem(hDlg, IDC_OUTPUT4);
			output5Wnd = GetDlgItem(hDlg, IDC_OUTPUT5);
			output6Wnd = GetDlgItem(hDlg, IDC_OUTPUT6);
			output7Wnd = GetDlgItem(hDlg, IDC_OUTPUT7);
			dutyCycleWnd = GetDlgItem(hDlg, IDC_DUTY_CYCLE);
			dutyCycleSliderWnd = GetDlgItem(hDlg, IDC_DUTY_CYCLE_SLIDER);
			lowTimeWnd = GetDlgItem(hDlg, IDC_LOW_TIME);
			highTimeWnd = GetDlgItem(hDlg, IDC_HIGH_TIME);

			// init controls
			data2gui(NULL);

			// init timer for every 100 ms
			SetTimer(hDlg, 1, 100, NULL);

			return (INT_PTR)TRUE;

		case WM_TIMER:
			if (connected) syncDevice();
			break;

		case WM_COMMAND:
			if (!inData2gui) {
				switch (LOWORD(wParam)) {
					case IDCANCEL:
						EndDialog(hDlg, LOWORD(wParam));
						return (INT_PTR)TRUE;
					case IDC_CONNECT:
						comOpen(comPort);
						connected = true;
						syncAll = true;
						data2gui(NULL);
						return (INT_PTR)TRUE;
					case IDC_DISCONNECT:
						connected = false;
						comClose();
						data2gui(NULL);
						return (INT_PTR)TRUE;
					case IDC_OFF:
						waveform = WAVEFORM_OFF;
						return (INT_PTR)TRUE;
					case IDC_SINE:
						waveform = WAVEFORM_SINE;
						return (INT_PTR)TRUE;
					case IDC_SQUARE_WAVE:
						waveform = WAVEFORM_SQUARE_WAVE;
						return (INT_PTR)TRUE;
					case IDC_NOISE:
						waveform = WAVEFORM_NOISE;
						return (INT_PTR)TRUE;
					case IDC_SAWTOOTH:
						waveform = WAVEFORM_SAWTOOTH;
						return (INT_PTR)TRUE;
					case IDC_INVERSE_SAWTOOTH:
						inverseSawtooth = isButtonChecked(inverseSawtoothWnd);
						return (INT_PTR)TRUE;
					case IDC_OUTPUT0:
						if (isButtonChecked(output0Wnd)) {
							digitalOutputBits |= 0x01;
						} else {
							digitalOutputBits &= ~0x01;
						}
						return (INT_PTR)TRUE;
					case IDC_OUTPUT1:
						if (isButtonChecked(output1Wnd)) {
							digitalOutputBits |= 0x02;
						} else {
							digitalOutputBits &= ~0x02;
						}
						return (INT_PTR)TRUE;
					case IDC_OUTPUT2:
						if (isButtonChecked(output2Wnd)) {
							digitalOutputBits |= 0x04;
						} else {
							digitalOutputBits &= ~0x04;
						}
						return (INT_PTR)TRUE;
					case IDC_OUTPUT3:
						if (isButtonChecked(output3Wnd)) {
							digitalOutputBits |= 0x08;
						} else {
							digitalOutputBits &= ~0x08;
						}
						return (INT_PTR)TRUE;
					case IDC_OUTPUT4:
						if (isButtonChecked(output4Wnd)) {
							digitalOutputBits |= 0x10;
						} else {
							digitalOutputBits &= ~0x10;
						}
						return (INT_PTR)TRUE;
					case IDC_OUTPUT5:
						if (isButtonChecked(output5Wnd)) {
							digitalOutputBits |= 0x20;
						} else {
							digitalOutputBits &= ~0x20;
						}
						return (INT_PTR)TRUE;
					case IDC_OUTPUT6:
						if (isButtonChecked(output6Wnd)) {
							digitalOutputBits |= 0x40;
						} else {
							digitalOutputBits &= ~0x40;
						}
						return (INT_PTR)TRUE;
					case IDC_OUTPUT7:
						if (isButtonChecked(output7Wnd)) {
							digitalOutputBits |= 0x80;
						} else {
							digitalOutputBits &= ~0x80;
						}
						return (INT_PTR)TRUE;
				}
				switch (HIWORD(wParam)) {
					case EN_KILLFOCUS:
						hwnd = (HWND) lParam;
						if (hwnd == frequencyStartWnd) {
							onFrequencyStartChanged();
						} else if (hwnd == frequencyWnd) {
							onFrequencyChanged();
						} else if (hwnd == frequencyEndWnd) {
							onFrequencyEndChanged();
						} else if (hwnd == periodStartWnd) {
							onPeriodStartChanged();
						} else if (hwnd == periodWnd) {
							onPeriodChanged();
						} else if (hwnd == periodEndWnd) {
							onPeriodEndChanged();
						} else if (hwnd == dutyCycleWnd) {
							onDutyCycleChanged();
						} else if (hwnd == lowTimeWnd) {
							onLowTimeChanged();
						} else if (hwnd == highTimeWnd) {
							onHighTimeChanged();
						}
						break;
				}
			}
			break;

		case WM_HSCROLL:
			if (!inData2gui) {
				hwnd = (HWND) lParam;
				if (hwnd == frequencySliderWnd) {
					onFrequencySliderChanged();
				} else if (hwnd == periodSliderWnd) {
					onPeriodSliderChanged();
				} else if (hwnd == dutyCycleSliderWnd) {
					onDutyCycleSliderChanged();
				}
			}
			break;
	}
	return (INT_PTR)FALSE;
}

int WINAPI WinMain(
	HINSTANCE hInstance,
    HINSTANCE hPrevInstance,
    LPSTR lpCmdLine,
    int nCmdShow)
{
	INITCOMMONCONTROLSEX init;
	InitCommonControlsEx(&init);
	DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAIN_DIALOG), NULL, mainDialog);
	return 0;
}

