-- Copyright (c) 2007 Frank Buss (fb@frank-buss.de)
--
-- Read and write register values:
-- RS232 port commands are evaluated:
-- read, example: "r 5". Reads register 5
-- write, example: "w 7 7f". Writes 0x7f to register 7
-- Every command is finished with a newline.
-- If readRequest is signalled, 2 ticks later the currentRegisterValue is
-- read, which has to be filled with the registerAddress register content.
-- If writeRequest is signalled, newRegisterValue contains the new value
-- for registerAddress.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity RegisterTransmitter is
    generic(systemSpeed, baudrate: integer);
    port(
        reset                : in std_logic;
        clock                : in std_logic;
        rx                   : in std_logic;
        tx                   : out std_logic;
        readRequest          : out std_logic;
        writeRequest         : out std_logic;
        registerAddress      : out unsigned(7 downto 0);
        newRegisterValue     : out unsigned(39 downto 0);
        currentRegisterValue : in unsigned(39 downto 0)
    );
end entity RegisterTransmitter;

architecture rtl of RegisterTransmitter is
    constant maxCounter: natural := systemSpeed / baudrate;
    
    subtype byte is unsigned(7 downto 0);
    subtype nibble is unsigned(3 downto 0);
    
    type stateType is (
        readRegisterAddressHigh,
        readRegisterAddressLow,
        readValue,
        readAddressWait,
        readAddress,
        sendNibble,
        waitForNibble,
        waitForNewline);
    
    signal state                : stateType := readRegisterAddressHigh;
    signal nibbleCounter        : natural range 0 to 9;
    signal address              : byte := (others => '0');
    signal value                : unsigned(39 downto 0) := (others => '0');

    signal rs232ReceivedByte    : byte := (others => '0');
    signal rs232NewByteReceived : std_logic := '0';
    signal rs232Data            : byte := (others => '0');
    signal rs232DataSendRequest : std_logic := '0';
    signal rs232DataSent        : std_logic := '0';
    
    -- convert an ASCII hex character to a byte
    function hexToUnsigned(ascii: byte) return nibble is
    begin
        if ascii >= x"30" and ascii <= x"39" then
            return ascii(3 downto 0);
        elsif ascii >= x"61" and ascii <= x"66" then
            return ascii(3 downto 0) + x"9";
        else
            return x"0";
        end if;
    end;
   
    -- convert a byte to an ASCII hex character
    function unsignedToHex(data: nibble) return byte is
    begin
        if data < x"a" then
            return x"30" + data;
        else
            return x"57" + data;
        end if;
    end;
   
begin
    
    instanceRS232Receiver: entity rs232Receiver
    generic map(
        systemSpeed     => systemSpeed,
        baudrate        => baudrate)
    port map(
        reset           => reset,
        clock           => clock,
        receivedByte    => rs232ReceivedByte,
        newByteReceived => rs232NewByteReceived,
        rx              => rx
    );

    instanceRS232Sender: entity rs232Sender
    generic map(
        systemSpeed     => systemSpeed,
        baudrate        => baudrate)
    port map(
        reset           => reset,
        clock           => clock,
        data            => rs232Data,
        tx              => tx,
        dataSendRequest => rs232DataSendRequest,
        dataSent        => rs232DataSent
    );
    
    process(clock, reset)
    begin
        if rising_edge(clock) then
            rs232DataSendRequest    <= '0';
            readRequest             <= '0';
            writeRequest            <= '0';
            if reset = '1' then
                state <= readRegisterAddressHigh;
            else
                case state is
                    when readRegisterAddressHigh =>
                        if rs232NewByteReceived = '1' then
                            if rs232ReceivedByte > 32 then
                                address <= x"0" & hexToUnsigned(rs232ReceivedByte);
                                state <= readRegisterAddressLow;
                            end if;
                        end if;
                    when readRegisterAddressLow =>
                        if rs232NewByteReceived = '1' then
                            if rs232ReceivedByte = x"3d" then
                                -- "=" received, try to receive new register value
                                value <= (others => '0');
                                state <= readValue;
                            else
                                if rs232ReceivedByte < 32 then
                                    -- newline or return received, read address
                                    registerAddress <= address;
                                    readRequest <= '1';
                                    state <= readAddressWait;
                                else
                                    -- next nibble of address
                                    address <= address(3 downto 0) & hexToUnsigned(rs232ReceivedByte);
                                end if;
                            end if;
                        end if;
                    when readValue =>
                        if rs232NewByteReceived = '1' then
                            if rs232ReceivedByte < 32 then
                                -- newline or return received, write new value
                                registerAddress <= address;
                                newRegisterValue <= value;
                                writeRequest <= '1';
                                state <= readRegisterAddressHigh;
                            else
                                -- next nibble of value
                                value <= value(35 downto 0) & hexToUnsigned(rs232ReceivedByte);
                            end if;
                        end if;
                    when readAddressWait =>
                        -- wait one more tick
                        state <= readAddress;
                    when readAddress =>
                        -- value available, read it
                        value <= currentRegisterValue;
                        nibbleCounter <= 9;
                        state <= sendNibble;
                    when sendNibble =>
                        -- send to RS232 port
                        rs232Data <= unsignedToHex(value(39 downto 36));
                        rs232DataSendRequest <= '1';
                        state <= waitForNibble;
                    when waitForNibble =>
                        -- wait until character sent
                        if rs232DataSent = '1' then
                            if nibbleCounter = 0 then
                                rs232Data <= x"0d";
                                rs232DataSendRequest <= '1';
                                state <= waitForNewline;
                            else
                                nibbleCounter <= nibbleCounter - 1;
                                value <= shift_left(value, 4);
                                state <= sendNibble;
                            end if;
                        end if;
                    when waitForNewline =>
                        if rs232DataSent = '1' then
                            state <= readRegisterAddressHigh;
                        end if;
                end case;
            end if;
        end if;
    end process;

end architecture rtl;
