-- Copyright (c) 2007 Frank Buss (fb@frank-buss.de)
--
-- RS232 receiver with fixed, but generic, baudrate for 8N1 mode.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity rs232Receiver is
    generic(systemSpeed, baudrate: integer);
    port(
        reset           : in std_logic;
        clock           : in std_logic;
        receivedByte    : out unsigned(7 downto 0);
        newByteReceived : out std_logic;
        rx              : in std_logic
    );
end entity rs232Receiver;

architecture rtl of rs232Receiver is
    constant maxCounter: natural := systemSpeed / baudrate;
    
    type stateType is (
        waitForRxStart, 
        waitHalfBit,
        receiveBits,
        waitForStopBit);
    
    signal state            : stateType := waitForRxStart;
    signal baudrateCounter  : natural range 0 to maxCounter := 0;
    signal bitCounter       : natural range 0 to 7 := 0;
    signal shiftRegister    : unsigned(7 downto 0) := (others => '0');
    signal rxLatched        : std_logic := '1';

begin

    process(clock, reset)
    begin
        if rising_edge(clock) then
            newByteReceived <= '0';
            if reset = '1' then
                state <= waitForRxStart;
                receivedByte <= (others => '0');
                newByteReceived <= '0';
                rxLatched <= '1';
            else
                rxLatched <= rx;
                case state is
                    when waitForRxStart =>
                        if rxLatched = '0' then
                            -- start bit received, wait for a half bit time
                            -- to sample bits in the middle of the signal
                            state <= waitHalfBit;
                            baudrateCounter <= maxCounter / 2 - 1;
                        end if;
                    when waitHalfBit =>
                        if baudrateCounter = 0 then
                            -- now we are in the middle of the start bit,
                            -- wait a full bit for the middle of the first bit
                            state <= receiveBits;
                            bitCounter <= 7;
                            baudrateCounter <= maxCounter - 1;
                        else
                            baudrateCounter <= baudrateCounter - 1;
                        end if;
                    when receiveBits =>
                        -- sample a bit
                        if baudrateCounter = 0 then
                            shiftRegister <= rxLatched & shiftRegister(7 downto 1);
                            if bitCounter = 0 then
                                state <= waitForStopBit;
                            else
                                bitCounter <= bitCounter - 1;
                            end if;
                            baudrateCounter <= maxCounter - 1;
                        else
                            baudrateCounter <= baudrateCounter - 1;
                        end if;
                    when waitForStopBit =>
                        -- wait for the middle of the stop bit
                        if baudrateCounter = 0 then
                            state <= waitForRxStart;
                            if rxLatched = '1' then
                                receivedByte <= shiftRegister;
                                newByteReceived <= '1';
                                -- else: missing stop bit, ignore
                            end if;    
                        else
                            baudrateCounter <= baudrateCounter - 1;
                        end if;
                end case;
            end if;
        end if;
    end process;

end architecture rtl;
