-- Copyright (c) 2007 Frank Buss (fb@frank-buss.de)
--
-- RS232 sender with fixed, but generic, baudrate for 8N1 mode.

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity rs232Sender is
    generic(
        systemSpeed,        -- clock speed, in hz
        baudrate: integer); -- baudrate, in bps
    port(
        clock               : in std_logic;
        reset               : in std_logic;
        data                : in unsigned(7 downto 0);
        tx                  : out std_logic;
        dataSendRequest     : in std_logic;
        dataSent            : out std_logic);
end entity rs232Sender;

architecture rtl of rs232Sender is
    constant maxCounter: natural := systemSpeed / baudrate;
    
    type stateType is (
        waitForStrobe,
        sendBits);

    signal state: stateType := waitForStrobe;

    signal baudrateCounter  : natural range 0 to maxCounter := 0;
    
    signal bitCounter       : natural range 0 to 10 := 0;
    signal shiftRegister    : unsigned(9 downto 0) := (others => '0');

begin

    update: process(clock)
    begin
        if rising_edge(clock) then
            dataSent <= '0';
            if reset = '1' then
                tx <= '1';
                state <= waitForStrobe;
            else
                case state is
                    -- wait until the master asserts valid data
                    when waitForStrobe =>
                        if dataSendRequest = '1' then
                            -- 1=stop bit, data, 0=start bit; data is sent LSB first
                            shiftRegister <= '1' & data & '0';
                            bitCounter <= 10;
                            baudrateCounter <= 0;
                            state <= sendBits;
                        else
                            tx <= '1';
                        end if;
                    when sendBits =>
                        if baudrateCounter = 0 then
                            baudrateCounter <= maxCounter - 1;
                            tx <= shiftRegister(0);
                            shiftRegister <= shift_right(shiftRegister, 1);
                            if bitCounter = 0 then
                                state <= waitForStrobe;
                                dataSent <= '1';
                            else
                                bitCounter <= bitCounter - 1;
                            end if;
                        else
                            baudrateCounter <= baudrateCounter - 1;
                        end if;
                end case;
            end if;
        end if;
    end process;
end architecture rtl;
